import os
import tempfile
import shutil

import gdal
import gdalconst


class PngTranslator:
    def __init__(self, png_path, x_origin, y_origin, x_dist, y_dist):
        self.png_path = png_path
        self.png = gdal.Open(png_path, gdal.GA_ReadOnly)

        self.x_origin = x_origin
        self.y_origin = y_origin
        self.x_dist = x_dist
        self.y_dist = y_dist

    def translate_to_tiff(self, output_path=None, **gdal_params):
        """[summary]
        PNGをTIFFに変換する

        Args:
            output_path (str, optional): TIFF出力先
            gdal_params (**dict): gdal_translateの追加パラメータ
        """
        # 一時フォルダを用意
        TMP_PATH = os.path.join(tempfile.gettempdir(), 'DemConverter')
        os.makedirs(TMP_PATH, exist_ok=True)

        # 入力PNGを一時フォルダにコピー
        png_name = os.path.basename(self.png_path)
        copy_png_path = os.path.join(TMP_PATH, png_name)
        shutil.copy(self.png_path, copy_png_path)

        # 一時PNG用のワールドファイルを生成してGDALでPNGを読み込み
        copy_pngw_path = copy_png_path + "w"
        self.make_worldfile(copy_pngw_path)
        copy_png = gdal.Open(copy_png_path, gdal.GA_ReadOnly)

        # 出力先が明示されていない場合一時フォルダに保存する
        if output_path is None:
            output_path = os.path.join(
                TMP_PATH, os.path.basename(self.png_path).split(".")[0] + ".tif")

        # convert 16bit greyscale PNG to Tiff
        gdal.Translate(output_path,
                       copy_png,
                       outputType=gdalconst.GDT_Float32,
                       **gdal_params)

        # TIF用のワールドファイルを生成
        self.make_worldfile(output_path[0:-4] + ".tfw")

        # 変換パラメータなどをCSVに保存
        self.make_csv(output_path[0:-4] + ".csv", {
            "output_type": "Float32",
            "xorigin": self.x_origin + self.x_dist / 2,
            "yorigin": self.y_origin - self.y_dist / 2,
            "xRes": self.x_dist,
            "yRes": self.y_dist,
            "src_min": gdal_params["scaleParams"][0][0],
            "src_max": gdal_params["scaleParams"][0][1],
            "dst_min": gdal_params["scaleParams"][0][2],
            "dst_max": gdal_params["scaleParams"][0][3],
            "a_srs": gdal_params["outputSRS"],
            "elev_lowest": gdal_params["scaleParams"][0][2],
            "elev_multiplier": (gdal_params["scaleParams"][0][3] - gdal_params["scaleParams"][0][2])
            / (gdal_params["scaleParams"][0][1] - gdal_params["scaleParams"][0][0])
        })

    def make_worldfile(self, worldfile_path: str):
        """[summary]
        ワールドファイルを生成する

        Args:
            pngw_path (str): ワールドファイル出力先
            x_origin (float): PNGの左上原点の地理座標x
            y_origin (float): PNGの左上原点の地理座標y
            x_res (float): PNGのx解像度
            y_res (float): PNGのy解像度
        """

        with open(worldfile_path, mode="w") as f:
            worldfile_str = ""
            worldfile_str += str(self.x_dist) + "\n"
            worldfile_str += "0.0\n"
            worldfile_str += "0.0\n"
            worldfile_str += str(self.y_dist * -1) + "\n"
            worldfile_str += str(self.x_origin + self.x_dist / 2) + "\n"
            worldfile_str += str(self.y_origin - self.y_dist / 2) + "\n"
            f.write(worldfile_str)

    def make_csv(self, csv_path: str, params: dict):
        import csv
        with open(csv_path, mode="w", encoding='shift_jis') as f:
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(params.keys())
            writer.writerow(params.values())
