import os
import tempfile

import gdal
import gdalconst

from .. import settings


class GeoTiffManager:
    def __init__(self, tiff_path):
        self.tiff_path = tiff_path
        self.tiff = gdal.Open(tiff_path, gdal.GA_ReadOnly)
        self.transform = self.tiff.GetGeoTransform()
        self.proj = self.tiff.GetProjection()

    def get_origin(self):
        return self.transform[0], self.transform[3]

    def get_distance_per_pixel(self):
        return self.transform[1], self.transform[5]

    def get_info(self, key=None):
        gdalinfo = gdal.Info(self.tiff, format='json')
        if key:
            return gdalinfo.get(key)

        return gdalinfo

    def get_tiff_min_max(self):
        stats = self.tiff.GetRasterBand(1).GetStatistics(True, True)
        return stats[0], stats[1]

    def calc_src_min_max(self, bUpperMarginAuto, lower_margin: float, upper_margin: float):
        """PNGへの変換パラメータとして、マージンと所定の標高幅を考慮した最低標高と最高標高を算出する

        Args:
            lower_margin (float): 標高の下方マージン

        Returns:
            (tuple): src_min, src_max(PNG=0の時の標高値, PNG=65535の時の標高値)
        """
        tiff_min, tiff_max = self.get_tiff_min_max()

        # 最低標高を算出
        src_min = tiff_min - round(lower_margin, settings.HEIGHT_MARGIN_DECIMAL_DIGITS)
        src_max = 0

        if bUpperMarginAuto:
            # 標高幅リストを小さい順に走査して、可能な限り小さい標高幅を用いて最高標高を算出する
            for diff in settings.HEIGHT_DIFF_LIST:
                if (tiff_max - src_min) > diff:
                    continue
                else:
                    src_max = diff + src_min
                    break
        else:
            src_max = tiff_max + round(upper_margin, settings.HEIGHT_MARGIN_DECIMAL_DIGITS)
        return round(src_min, settings.HEIGHT_DECIMAL_DIGITS), round(src_max, settings.HEIGHT_DECIMAL_DIGITS)

    def translate_to_png(self, output_path=None, bUpperMarginAuto=True, lower_margin=0.0, upper_margin=1000000.0, **gdal_params):
        if output_path is None:
            # 出力先が明示されていない場合一時フォルダを作成してそこにPNGを保存する
            TMP_PATH = os.path.join(tempfile.gettempdir(), 'DemConverter')
            os.makedirs(TMP_PATH, exist_ok=True)
            output_path = os.path.join(
                TMP_PATH, os.path.basename(self.tiff_path).split(".")[0] + ".png")

        src_min, src_max = self.calc_src_min_max(bUpperMarginAuto, lower_margin, upper_margin)
        dst_min, dst_max = 0, 65535
        scale_params = [[src_min, src_max, dst_min, dst_max]]

        # convert TIFF to 16bit greyscale PNG
        gdal.Translate(output_path,
                       self.tiff,
                       outputType=gdalconst.GDT_UInt16,
                       scaleParams=scale_params,
                       resampleAlg=self.get_resampleAlg(
                           gdal_params["xRes"], gdal_params["yRes"]),
                       noData=0,
                       ** gdal_params)

        return {
            "path": output_path,
            "info": gdal.Info(output_path, format="json"),
            "params_for_csv": {
                **gdal_params,
                "src_min": src_min,
                "src_max": src_max,
                "dst_min": dst_min,
                "dst_max": dst_max,
                "elev_lowest": src_min,
                "elev_multiplier": (src_max - src_min) / (dst_max - dst_min)
            }
        }

    def get_resampleAlg(self, xRes: float, yRes: float) -> str:
        """
        リサンプリング手法を判定する
        入力のGeoTiffと変換先のx,y地理解像度に変更がないならnearest
        変更があるなら（スケーリングするなら）lanczos

        Args:
            xRes (float): 変換先のx解像度
            yRes (float): 変換先のy解像度

        Returns:
            str: resampleAlg文字列
        """
        src_xRes, src_yRes = self.get_distance_per_pixel()
        if src_xRes == xRes and src_yRes * -1 == yRes:
            return "nearest"
        else:
            return "lanczos"
