import os

from PIL import Image, ImageOps

from .. import settings


class PngProcessor:
    def __init__(self, png_path: str, origins: dict, x_dist: float, y_dist: float, quadrant: int):
        """
        PNGをタイル化するクラス

        Args:
            png_path (str): タイル化するPNGへのパス
            origins (dict): PNGのgdalinfoのうちcornerの情報
            x_dist (float): PNG1ピクセルあたりのx方向地理的距離
            y_dist (float): PNG1ピクセルあたりのy方向地理的距離
            quadrant (int): 矩形選択時の開始点からみた終了点が属する象限
        """
        self.png_path = png_path

        self.origins = origins
        self.x_dist = x_dist
        self.y_dist = y_dist

        self.quadrant = quadrant

        # PILによる画像切り出しは左上を原点として扱うのでPNGを上下左右反転させて対応
        Image.MAX_IMAGE_PIXELS = None         # これで制限を無効化します
        self.png = self.flip_and_mirror_by_quadrant(Image.open(png_path))

    def make_tile(self, x_res: float, y_res: float, x_space: float, y_space: float, output_dir: str):
        """
        タイルを生成する

        Args:
            x_res (float): タイル１枚あたりのx解像度
            y_res (float): タイル１枚あたりのy解像度
            output_dir (str): タイル出力先ディレクトリ
        """
        # タイル数の計算
        x_res = x_res - x_space*2
        y_res = y_res - y_space*2

        width, height = self.png.size
        x_rows = (width-x_space*2) // x_res
        y_cols = (height-y_space*2) // y_res

        for y in range(y_cols):
            for x in range(x_rows):
                # 指定解像度で画像を切り抜く
                cropped = self.png.crop((x * x_res, y * y_res,
                                         (x + 1) * x_res + x_space*2, (y + 1) * y_res + y_space*2))
                # クラス初期化時の反転を元に戻す
                unflipped = self.flip_and_mirror_by_quadrant(cropped)
                # 画像を書き出す
                unflipped.save(os.path.join(output_dir, f"{x}-{y}.png"))
                self.make_worldfile(x_res, y_res, x_space, y_space, x, y, output_dir)

    def make_worldfile(self,
                       x_res: float, y_res: float,
                       x_space: float, y_space: float,
                       x_count: int, y_count: int,
                       output_dir: str):
        """[summary]
        PNGのワールドファイルをタイルごとに生成する

        Args:
            x_res (float): PNGのx解像度
            y_res (float): PNGのy解像度
            x_count (int): 左上を(x, y)=(0, 0)としたときのタイルのx番地
            y_count (int): 左上を(x, y)=(0, 0)としたときのタイルのy番地
            output_dir (str): 出力先ディレクトリ
        """

        # 矩形の選択方向に応じて変数を計算
        x_operand, y_operand = self.get_xy_operand_by_quadrant()
        x_offset, y_offset = self.get_xy_offset_by_quadrant()
        x_origin, y_origin = self.get_xy_origin_by_quadrant()

        # タイル化されたPNGの地理的原点を計算
        # xの地理的原点 = 元のPNGの地理的原点 + タイルのx番地 * x解像度 * 1ピクセルあたり地理的距離 + オフセット値
        # オフセット値: 上記の式は終了点が第4象限にある場合の原点についてで、他の象限の場合に一定距離ズラす値のこと
        offset_x_origin = x_origin + (x_count * x_res - x_space) * \
            self.x_dist * x_operand + x_offset * x_res * self.x_dist
        offset_y_origin = y_origin + (y_count * y_res - y_space) * \
            self.y_dist * y_operand + y_offset * y_res * self.y_dist

        with open(os.path.join(output_dir, f"{x_count}-{y_count}.pngw"), mode="w") as f:
            worldfile_str = ""
            worldfile_str += str(round(self.x_dist,
                                       settings.PX_DIST_DECIMAL_DIGITS)) + "\n"
            worldfile_str += "0.0\n"
            worldfile_str += "0.0\n"
            worldfile_str += str(round(self.y_dist * -1,
                                       settings.PX_DIST_DECIMAL_DIGITS)) + "\n"
            worldfile_str += str(offset_x_origin + self.x_dist / 2) + "\n"
            worldfile_str += str(offset_y_origin - self.y_dist / 2) + "\n"
            f.write(worldfile_str)

    def flip_and_mirror_by_quadrant(self, img) -> Image:
        """
        象限ごとに画像を上下左右反転させる

        Args:
            img (PIL.Image): Imageのインスタンス

        Returns:
            Image: 上下左右反転されたImageのインスタンス
        """
        if self.quadrant == 1:
            return ImageOps.flip(img)
        elif self.quadrant == 2:
            return ImageOps.mirror(ImageOps.flip(img))
        elif self.quadrant == 3:
            return ImageOps.mirror(img)
        else:
            return img

    def get_xy_operand_by_quadrant(self):
        """
        象限ごとにx, y座標計算上の正負符号を返す

        Returns:
            (int, int): (x, y)それぞれの正負符号
        """
        if self.quadrant == 1:
            return (1, 1)
        elif self.quadrant == 2:
            return (-1, 1)
        elif self.quadrant == 3:
            return (-1, -1)
        else:
            return (1, -1)

    def get_xy_offset_by_quadrant(self):
        """
        象限ごとにタイル切り出し開始位置を計算するためのオフセット値を返す

        Returns:
            (int, int): (x, y)それぞれのオフセット値
        """
        if self.quadrant == 1:
            return (0, 1)
        elif self.quadrant == 2:
            return (-1, 1)
        elif self.quadrant == 3:
            return (-1, 0)
        else:
            return (0, 0)

    def get_xy_origin_by_quadrant(self):
        """
        象限ごとに、切り出したタイルの四隅いずれをワールドファイル上の原点とするか判定し値を返す

        Returns:
            (float, float): (x, y)それぞれの原点
        """
        if self.quadrant == 1:
            xy_origin = self.origins.get("lowerLeft")
            return (xy_origin[0], xy_origin[1])
        elif self.quadrant == 2:
            xy_origin = self.origins.get("lowerRight")
            return (xy_origin[0], xy_origin[1])
        elif self.quadrant == 3:
            xy_origin = self.origins.get("upperRight")
            return (xy_origin[0], xy_origin[1])
        else:
            xy_origin = self.origins.get("upperLeft")
            return (xy_origin[0], xy_origin[1])
